import javafx.geometry.Pos;
import org.junit.Assert;

import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/*
Schreiben bzw. vervollständigen Sie gekennzeichnete Programmteile wie in den Kommentaren beschrieben.
Verändern Sie aber nicht vorgegebene Methodenköpfe oder andere vorgegebene Programmteile!
Verwenden Sie die gegebenen Testklassen zum Überprüfen Ihrer Lösung. Die Korrektheit dieser Testfälle
ist notwendig, aber nicht hinreichend für eine korrekte Lösung. Es werden damit nur bestimmte
Spezialfälle geprüft. Liefern die Testklassen keinen Fehler, garantiert dies noch nicht die
Korrektheit Ihrer Lösung gemäß der Angabe.
*/

public class Aufgabe1 {

    /*
    Aufgabe: Umgang mit Klassen und Objekten, sowie vorgefertigten Datenstrukturen.

        Ergänzen Sie fehlende Teile der Klassen 'Position' und 'Image' entsprechend den Kommentaren
        an den mit TODO gekennzeichneten Stellen.

    Punkte (maximal 40):
        10 Punkte in Summe für Klasse 'Position':
            1 Punkt  für den Konstruktor,
            3 Punkte für 'hasSmallerX' und 'hasSmallerY'
            3 Punkte für 'toString',
            3 Punkte für 'equals',
        30 Punkte in Summe für Klasse 'Image':
            6 Punkte für den Konstruktor
            6 Punkte für beide 'add'-Methoden
            6 Punkte für 'selectRange'
            6 Punkte für 'copy'
            6 Punkte für 'toString'

     Auch für teilweise korrekte Lösungen werden Punkte vergeben.

    */


    // Diese Methode können Sie zum Testen nutzen. Sie geht nicht in die Beurteilung ein.
    public static void main(String[] args) {

        //Testen von Position:
        Position p = new Position(1, 2);
        System.out.println(p);
        System.out.println();

        for (int x = 0; x < 5; x++) {
            for (int y = 0; y < 5; y++) {
                if ((x == 1 && y == 2) != new Position(x, y).equals(p)) {
                    System.out.println("Fehler in Klasse 'Position'!");
                }
            }
        }

        //Testen von Image:
        Image image = new Image(3, 3);

        image.add(0, 0, 'I');
        image.add(1, 1, '#');
        image.add(new Position(2, 2), '#');
        image.add(new Position(0, 1), '*');
        image.add(new Position(2, 1), '+');

        List<Position> list = image.selectRange(1, 2);

        System.out.println(list);
        System.out.println();

        System.out.println(image);

        Image cropped = image.copy(2, 2);
        System.out.println(cropped);

        System.out.println(image);
        /* Erwartete Ausgabe:

        (1,2)

        [(0,0), (0,1)]

        I..
        *#+
        ..#

        I.
        *#

        I..
        *#+
        ..#

        */
    }
}

// Diese Klasse repräsentiert einen unveränderbaren 2D-Punkt mit 2 Koordinaten in der Ebene.
class Position {

    private final int x;
    private final int y;

    // Konstruktor: initialisiert diesen Punkt mit seinen Koordinaten.
    // Wenn 'x' oder 'y' kleiner 0 ist, wird der entsprechende Wert 0 gesetzt.
    public Position(int x, int y) {
        this.x = x < 0 ? 0 : x;
        this.y = y < 0 ? 0 : y;
    }

    // Liefert true, genau dann wenn die x-Kordinate 'x' kleiner als die von 'p' ist
    public boolean hasSmallerX(Position p) {
        return x < p.x;
    }

    // Liefert true, genau dann wenn die y-Kordinate 'y' kleiner als die von 'p' ist
    public boolean hasSmallerY(Position p) {
        return y < p.y;
    }

    @Override
    // Liefert eine lesbare Repräsentation der Koordinaten dieser Position in
    // der Form:
    // (x,y)
    public String toString() {
        return "(" + x + "," + y + ")";
    }

    @Override
    // Die equals-Methode: Überprüft die Gleichheit im Bezug auf
    // Koordinaten mit einem angegebenen Objekt.
    public boolean equals(Object o) {
        if (o == null) {
            return false;
        }
        if (o.getClass() != Position.class) {
            return false;
        }
        Position p = (Position) o;
        return x == p.x && y == p.y;

    }

    @Override
    // Die hashCode-Methode, die den Hashcode dieses Punktes liefert.
    // Diese Methode nicht ändern.
    public int hashCode() {
        return this.x + this.y;

    }
}

// Diese Klasse repräsentiert ein rechteckiges Bild mit
// 'height' Zeilen der Länge 'width'.
// In das Bild lassen sich an Positionen Zeichen setzen.
class Image {

    // Deklarieren Sie keine weiteren Objekt-/globalen Variablen.

    // Speichert Positionen mit den gesetzten Zeichen.
    private HashMap<Position, Character> map;
    // Höhe und Breite des Bildes (gültiger Bildbereich von 0 bis height-1 bzw. width-1).
    private int height, width;

    // Konstruktor: initialisiert dieses 'Image' als leeres Bild ohne
    // gesetzte Punkte.
    // Annahme: width > 0 und height > 0
    public Image(int width, int height) {
        this.width = width;
        this.height = height;

        map = new HashMap<>();
    }

    // Setzt das Zeichen 'ch' an der Position 'position', sofern die
    // Position im Bildbereich liegt (0 <= x < width und 0 <= y < height).
    // Wenn der Punkt bereits gesetzt war, wird das Zeichen ersetzt.
    // Liegt 'position' außerhalb des Bildbereichs,
    // wird der Punkt nicht gesetzt und die Methode liefert 'false'.
    // Liefert true, genau dann wenn das Zeichen gesetzt
    // wird.
    // Annahme: position != null
    public boolean add(Position position, char ch) {
        Position minPosition = new Position(0, 0);
        Position maxPosition = new Position(width, height);

        if (position.hasSmallerX(minPosition) || position.hasSmallerY(minPosition)) return false;

        if (position.hasSmallerX(maxPosition) && position.hasSmallerY(maxPosition)) {
            map.put(position, ch);
            return true;
        }

        return false;
    }

    // Macht dasselbe wie 'add(Position, char)', allerdings werden die Koordinaten
    // des hinzuzufügenden Zeichens separat angegeben.
    // Nutzen Sie zur Implementierung 'add(Position, char)'.
    public boolean add(int x, int y, char ch) {
        return add(new Position(x, y), ch);
    }


    // Wählt unter allen gesetzten Positionen (x,y) jene aus, die im
    // angegebenen Bereich 0 <= x < selectedWidth und 0 <= y < selectedHeight liegen.
    // Liefert die ausgewählten Positionen (dort wo Zeichen stehen) in einer Liste zurück.
    // Gibt es in dem Bereich keine gesetzten Positionen, ist die Liste leer.
    // Annahme: selectedWidth >= 0, selectedHeight >= 0
    public LinkedList<Position> selectRange(int selectedWidth, int selectedHeight) {
        LinkedList<Position> positions = new LinkedList<>();
        Position selectedPosition = new Position(selectedWidth, selectedHeight);

        for (Position position : map.keySet()) {
            if (position.hasSmallerX(selectedPosition) && position.hasSmallerY(selectedPosition)) {
                positions.offer(position);
            }
        }

        return positions;
    }

    // Liefert einen angegebenen rechteckigen Bereich mit Positionen (x,y),
    // wobei 0 <= x < selectedWidth und 0 <= y < selectedHeight
    // als neues Bild zurück.
    public Image copy(int selectedWidth, int selectedHeight) {
        Image copy = new Image(selectedWidth, selectedHeight);

        Position minPosition = new Position(0, 0);
        Position maxPosition = new Position(selectedWidth, selectedHeight);

        HashMap<Position, Character> copyMap = new HashMap<>();

        for (Position position : map.keySet()) {
            if (position.hasSmallerX(minPosition) || position.hasSmallerY(minPosition)) continue;
            if (position.hasSmallerX(maxPosition) && position.hasSmallerY(maxPosition)) {
                copyMap.put(position, map.get(position));
            }
        }

        copy.map = copyMap;

        return copy;

    }

    @Override
    // Liefert eine lesbare Repräsentation, in der das Bild
    // als Raster mit 'height' Zeilen der Länge 'width' dargestellt wird.
    // Dort, wo keine Zeichen gesetzt sind, steht ein '.'
    // (siehe Beispiele in 'main').
    public String toString() {
        String output = "";

        for (int h = 0; h < height; h++) {
            for (int w = 0; w < width; w++) {
                Position position = new Position(w, h);
                if (map.containsKey(position)) {
                    output += map.get(position);
                } else {
                    output += ".";
                }
            }
            output += "\n";
        }
        return output;
    }

}
